@@ -2,7 +2,7 @@
 # Modified by Cameron Dale
 # see LICENSE.txt for license information
 #
-# $Id: track.py 280 2007-08-22 20:11:14Z camrdale-guest $
+# $Id: track.py 365 2008-01-28 04:27:31Z camrdale-guest $
 
 """Tools to track a download swarm.
 
@@ -160,12 +160,21 @@ def statefiletemplate(x):
                         raise ValueError
                     if type(info.get('requirecrypto')) not in (IntType,LongType):
                         raise ValueError
-        elif cname == 'completed':
-            if (type(cinfo) != DictType): # The 'completed' key is a dictionary of SHA hashes (torrent ids)
+        elif cname == 'stats':
+            if (type(cinfo) != DictType): # The 'stats' key is a dictionary of SHA hashes (torrent ids)
                 raise ValueError          # ... for keeping track of the total completions per torrent
             for y in cinfo.values():      # ... each torrent has an integer value
-                if type(y) not in (IntType,LongType):
+                if type(y) != DictType:
                     raise ValueError      # ... for the number of reported completions for that torrent
+                completed = y.get('completed')
+                if type(completed) not in (IntType,LongType) or completed < 0:
+                    raise ValueError
+                uploaded = y.get('uploaded', 0)
+                if type(uploaded) not in (IntType,LongType) or uploaded < 0:
+                    raise ValueError
+                downloaded = y.get('downloaded', 0)
+                if type(downloaded) not in (IntType,LongType) or downloaded < 0:
+                    raise ValueError
         elif cname == 'allowed':
             if (type(cinfo) != DictType): # a list of info_hashes and included data
                 raise ValueError
@@ -186,7 +195,50 @@ def statefiletemplate(x):
                 if dirkeys.has_key(y[1]): # and each should have a unique info_hash
                     raise ValueError
                 dirkeys[y[1]] = 1
-            
+        elif cname == 'version':
+            if (type(cinfo) != IntType): # the version of the state file
+                raise ValueError
+
+def loadState(configdir):
+    """Load the saved state from a previous run.
+    
+    @type configdir: L{DebTorrent.ConfigDir.ConfigDir}
+    @param configdir: the configuration and cache directory manager
+    @rtype: C{dictionary}
+    @return: the saved state from the previous run
+    
+    """
+    
+    tempstate = configdir.getState()
+    if tempstate:
+        version = tempstate.get('version', 0)
+        # Upgrade from the previous version
+        if version < 1:
+            try:
+                tempstate['stats'] = {}
+                if 'completed' in tempstate:
+                    for hash, num in tempstate['completed'].items():
+                        tempstate['stats'][hash] = {'completed': num,
+                                                    'downloaded': 0,
+                                                    'uploaded': 0}
+                    del tempstate['completed']
+                tempstate['version'] = 1
+                logger.info('successfully upgraded the previous state file')
+            except:
+                logger.exception('old statefile corrupt; resetting')
+                tempstate = {'version': 1}
+
+        # Test the state to make sure it's not been corrupted
+        try:
+            statefiletemplate(tempstate)
+            logger.info('successfully loaded the previous state file')
+        except:
+            logger.exception('statefile corrupt; resetting')
+            tempstate = {'version': 1}
+    else:
+        tempstate = {'version': 1}
+    
+    return tempstate
 
 alas = 'your file may exist elsewhere in the universe\nbut alas, not here\n'
 
@@ -283,6 +335,36 @@ def compact_peer_info(ip, port):
         s = ''  # not a valid IP, must be a domain name
     return s
 
+def clean_name(name):
+    """Clean up the name of the torrent to remove mirror info.
+    
+    @type name: C{string}
+    @param name: the torrent name
+    @rtype: C{string}
+    @return: the cleaned name
+    
+    """
+    
+    if not name:
+        return ''
+    
+    safe_name = name
+    try:
+        # remove the mirror name
+        safe_name = safe_name.split('_', 1)[1]
+    except:
+        pass
+    try:
+        # remove everything to the last 'dists'
+        safe_name = safe_name[safe_name.rindex('dists_'):]
+    except:
+        try:
+            # remove everything to the last 'debian'
+            safe_name = safe_name[safe_name.rindex('debian_')+7:]
+        except:
+            pass
+    return safe_name
+
 class Tracker:
     """Track a download swarm.
     
@@ -329,9 +411,9 @@ class Tracker:
     @ivar downloads: keys are info hashes, values are dictionaries with keys the
         peer IDs of peers connected to that torrent and values the dictionaries
         of information about the peer
-    @type completed: C{dictionary}
-    @ivar completed: keys are info hashes, values are the number of peers that
-        have completed downloading the entire torrent
+    @type stats: C{dictionary}
+    @ivar stats: keys are info hashes, values are dictionaries containing
+        statistics related to the torrent
     @type becache: C{list} of C{list} of C{dictionary}
     @ivar becache: keys are the infohashes, values are the cached peer data.
         peer set format::
@@ -425,8 +507,8 @@ class Tracker:
         self.cached = {}    # format: infohash: [[time1, l1, s1], [time2, l2, s2], ...]
         self.cached_t = {}  # format: infohash: [time, cache]
         self.times = {}
-        self.state = {}
         self.seedcount = {}
+        self.torrent_names = {}
 
         self.allowed_IPs = None
         self.banned_IPs = None
@@ -443,18 +525,9 @@ class Tracker:
             logger.warning('crypto library not installed,' +
                    ' cannot completely verify encrypted peers')
 
-        tempstate = configdir.getState()
-        if tempstate:
-            try:
-                if not tempstate.has_key('peers'):
-                    tempstate = {'peers': tempstate}
-                statefiletemplate(tempstate)
-                self.state = tempstate
-                logger.info('successfully loaded the previous state file')
-            except:
-                logger.warning('statefile corrupt; resetting')
+        self.state = loadState(configdir)
         self.downloads = self.state.setdefault('peers', {})
-        self.completed = self.state.setdefault('completed', {})
+        self.stats = self.state.setdefault('stats', {})
 
         self.becache = {}
 
@@ -463,6 +536,8 @@ class Tracker:
         else:
             self.cache_default_len = 5
         for infohash, ds in self.downloads.items():
+            torrent_name = []
+            self.torrent_names[infohash] = torrent_name
             self.seedcount[infohash] = 0
             for x,y in ds.items():
                 ip = y['ip']
@@ -472,6 +547,9 @@ class Tracker:
                     continue
                 if not y['left']:
                     self.seedcount[infohash] += 1
+                name = clean_name(y.get('name', ''))
+                if name and name not in torrent_name:
+                    torrent_name.append(name)
                 if y.get('nat',-1):
                     continue
                 gip = y.get('given_ip')
@@ -619,17 +697,21 @@ class Tracker:
                 return (302, 'Found', {'Content-Type': 'text/html', 'Location': red},
                         '<A HREF="'+red+'">Click Here</A>')
             
+            # Write the document header
             s = StringIO()
-            s.write('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n' \
-                '<html><head><title>DebTorrent download info</title>\n')
+            s.write('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" ' \
+                    '"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n' \
+                    '<html><head><title>DebTorrent tracker info</title>\n')
             if self.favicon is not None:
                 s.write('<link rel="shortcut icon" href="/favicon.ico">\n')
             s.write('</head>\n<body>\n' \
-                '<h3>DebTorrent download info</h3>\n'\
-                '<ul>\n'
-                '<li><strong>tracker version:</strong> %s</li>\n' \
-                '<li><strong>server time:</strong> %s</li>\n' \
-                '</ul>\n' % (version, isotime()))
+                    '<h3>DebTorrent tracker info</h3>\n'\
+                    '<ul>\n'
+                    '<li><strong>tracker version:</strong> %s</li>\n' \
+                    '<li><strong>server time:</strong> %s</li>\n' \
+                    '</ul>\n' % (version, isotime()))
+
+            # Collect the torrent hashes and names
             if self.config['allowed_dir']:
                 if self.show_names:
                     names = [ (self.allowed[hash]['name'],hash)
@@ -638,64 +720,102 @@ class Tracker:
                     names = [ (None,hash)
                               for hash in self.allowed.keys() ]
             else:
-                names = [ (None,hash) for hash in self.downloads.keys() ]
+                if self.show_names:
+                    names = [ ('\n<br>'.join(self.torrent_names[hash]),hash) for hash in self.downloads.keys() ]
+                else:
+                    names = [ (None,hash) for hash in self.downloads.keys() ]
+
             if not names:
                 s.write('<p>not tracking any files yet...</p>\n')
             else:
                 names.sort()
-                tn = 0
-                tc = 0
-                td = 0
-                tt = 0  # Total transferred
-                ts = 0  # Total size
-                nf = 0  # Number of files displayed
-                if self.config['allowed_dir'] and self.show_names:
-                    s.write('<table summary="files" border="1">\n' \
-                        '<tr><th>info hash</th><th>torrent name</th><th align="right">size</th><th align="right">complete</th><th align="right">downloading</th><th align="right">downloaded</th><th align="right">transferred</th></tr>\n')
-                else:
-                    s.write('<table summary="files">\n' \
-                        '<tr><th>info hash</th><th align="right">complete</th><th align="right">downloading</th><th align="right">downloaded</th></tr>\n')
+                number_files = 0
+                total_peers = 0
+                total_uploaded = 0L
+                total_downloaded = 0L
+                total_size = 0L
+                
+                # Write the table headers
+                s.write('<table summary="files" border="1">\n\n' \
+                        '<tr><th>torrent name/<br><code>    info hash</code></th>\n')
+                if self.config['allowed_dir']:
+                    s.write('<th align="right">size</th>\n')
+                s.write('<th align="right">peers</th>\n' \
+                        '<th align="right">downloaded</th>\n' \
+                        '<th align="right">uploaded</th>\n' \
+                        '<th align="right">saved</th></tr>\n\n')
+
+                # Display a table row for each torrent
                 for name,hash in names:
-                    l = self.downloads[hash]
-                    n = self.completed.get(hash, 0)
-                    tn = tn + n
-                    c = self.seedcount[hash]
-                    tc = tc + c
-                    d = len(l) - c
-                    td = td + d
-                    if self.config['allowed_dir'] and self.show_names:
-                        if self.allowed.has_key(hash):
-                            nf = nf + 1
-                            sz = self.allowed[hash]['length']  # size
-                            ts = ts + sz
-                            szt = sz * n   # Transferred for this torrent
-                            tt = tt + szt
-                            if self.allow_get == 1:
-                                linkname = '<a href="/file?info_hash=' + quote(hash) + '">' + name + '</a>'
-                            else:
-                                linkname = name
-                            s.write('<tr><td><code>%s</code></td><td>%s</td><td align="right">%s</td><td align="right">%i</td><td align="right">%i</td><td align="right">%i</td><td align="right">%s</td></tr>\n' \
-                                % (b2a_hex(hash), linkname, size_format(sz), c, d, n, size_format(szt)))
+                    # Get the stats for this torrent
+                    uploaded = self.stats.get(hash, {}).get('uploaded', 0)
+                    total_uploaded += uploaded
+                    downloaded = self.stats.get(hash, {}).get('downloaded', 0)
+                    total_downloaded += downloaded
+                    saved = 0.0
+                    if downloaded > 0:
+                        saved = 100.0 * float(uploaded) / float(downloaded)
+                    peers = len(self.downloads[hash])
+                    total_peers += peers
+
+                    # Display is different if we are using allowed_dir
+                    if self.config['allowed_dir'] and self.allowed.has_key(hash):
+                        number_files += 1
+                        size = self.allowed[hash]['length']  # size
+                        total_size += long(size)
+                        if self.allow_get == 1:
+                            linkname = '<a href="/file?info_hash=' + quote(hash) + '">' + name + '</a>'
+                        else:
+                            linkname = name
+                        s.write('<tr><td>%s<br><code>    %s</code></td>\n' \
+                                '<td align="right">%s</td>\n' \
+                                '<td align="right">%i</td>\n' \
+                                '<td align="right">%s</td>\n' \
+                                '<td align="right">%s</td>\n' \
+                                '<td align="right">%0.1f %%</td></tr>\n\n' \
+                                % (linkname, b2a_hex(hash), size_format(size),
+                                   peers, size_format(downloaded),
+                                   size_format(uploaded), saved))
                     else:
-                        s.write('<tr><td><code>%s</code></td><td align="right"><code>%i</code></td><td align="right"><code>%i</code></td><td align="right"><code>%i</code></td></tr>\n' \
-                            % (b2a_hex(hash), c, d, n))
-                if self.config['allowed_dir'] and self.show_names:
-                    s.write('<tr><td align="right" colspan="2">%i files</td><td align="right">%s</td><td align="right">%i</td><td align="right">%i</td><td align="right">%i</td><td align="right">%s</td></tr>\n'
-                            % (nf, size_format(ts), tc, td, tn, size_format(tt)))
-                else:
-                    s.write('<tr><td align="right">%i files</td><td align="right">%i</td><td align="right">%i</td><td align="right">%i</td></tr>\n'
-                            % (nf, tc, td, tn))
+                        number_files += 1
+                        s.write('<tr><td>')
+                        if name:
+                            s.write('%s<br>' % name)
+                        s.write('<code>    %s</code></td>\n' \
+                                '<td align="right">%i</td>\n' \
+                                '<td align="right">%s</td>\n' \
+                                '<td align="right">%s</td>\n' \
+                                '<td align="right">%0.1f %%</td></tr>\n\n' \
+                                % (b2a_hex(hash), peers, size_format(downloaded),
+                                   size_format(uploaded), saved))
+
+                # The table footer with summary statistics
+                saved = 0.0
+                if total_downloaded > 0:
+                    saved = 100.0 * float(total_uploaded) / float(total_downloaded)
+                s.write('<tr><td align="right">%i torrents</td>\n' % number_files)
+                if self.config['allowed_dir']:
+                    s.write('<td align="right">%s</td>\n' % size_format(total_size))
+                s.write('<td align="right">%i</td>\n' \
+                        '<td align="right">%s</td>\n' \
+                        '<td align="right">%s</td>\n' \
+                        '<td align="right">%0.1f %%</td></tr>\n\n'
+                        % (total_peers, size_format(total_downloaded),
+                           size_format(total_uploaded), saved))
+
+                # Write a rudimentary legend
                 s.write('</table>\n' \
                     '<ul>\n' \
-                    '<li><em>info hash:</em> SHA1 hash of the "info" section of the metainfo (*.dtorrent)</li>\n' \
-                    '<li><em>complete:</em> number of connected clients with the complete file</li>\n' \
-                    '<li><em>downloading:</em> number of connected clients still downloading</li>\n' \
-                    '<li><em>downloaded:</em> reported complete downloads</li>\n' \
-                    '<li><em>transferred:</em> torrent size * total downloaded (does not include partial transfers)</li>\n' \
+                    '<li><em>torrent name:</em> possible names for the torrent</li>\n' \
+                    '<li><em>info hash:</em> torrent identifier</li>\n' \
+                    '<li><em>peers:</em> number of connected clients</li>\n' \
+                    '<li><em>downloaded:</em> amount of data downloaded by all peers, this data may have been downloaded from a mirror or another peer</li>\n' \
+                    '<li><em>uploaded:</em> amount of data uploaded by all peers, this is the amount of bandwidth that was saved from the mirrors</li>\n' \
+                    '<li><em>saved:</em> percentage of data that was NOT downloaded from a mirror</li>\n' \
                     '</ul>\n')
 
             s.write('</body>\n' \
-                '</html>\n')
+                    '</html>\n')
             return (200, 'OK', {'Content-Type': 'text/html; charset=iso-8859-1'}, s.getvalue())
         except:
             logger.exception('Error generating info page')
@@ -716,7 +836,7 @@ class Tracker:
         """
         
         l = self.downloads[hash]
-        n = self.completed.get(hash, 0)
+        n = self.stats.get(hash, {}).get('completed', 0)
         c = self.seedcount[hash]
         d = len(l) - c
         f = {'complete': c, 'incomplete': d, 'downloaded': n}
@@ -854,8 +974,11 @@ class Tracker:
         
         peers = self.downloads.setdefault(infohash, {})
         ts = self.times.setdefault(infohash, {})
-        self.completed.setdefault(infohash, 0)
+        stats = self.stats.setdefault(infohash, {'completed': 0,
+                                                 'uploaded': 0L,
+                                                 'downloaded': 0L})
         self.seedcount.setdefault(infohash, 0)
+        torrent_name = self.torrent_names.setdefault(infohash, [])
 
         def params(key, default = None, l = paramslist):
             """Get the user parameter, or the default.
@@ -887,11 +1010,12 @@ class Tracker:
         port = long(port)
         if port < 0 or port > 65535:
             raise ValueError, 'invalid port'
+        name = params('name')
         left = long(params('left',''))
         if left < 0:
             raise ValueError, 'invalid amount left'
-        uploaded = long(params('uploaded',''))
-        downloaded = long(params('downloaded',''))
+        uploaded = long(params('uploaded', 0L))
+        downloaded = long(params('downloaded', 0L))
         if params('supportcrypto'):
             supportcrypto = 1
             try:
@@ -923,18 +1047,31 @@ class Tracker:
 
         if event == 'stopped':
             if peer:
+                if uploaded > peer.setdefault('uploaded', uploaded):
+                    stats['uploaded'] += uploaded - peer['uploaded']
+                    peer['uploaded'] = uploaded
+                if downloaded > peer.setdefault('downloaded', downloaded):
+                    stats['downloaded'] += downloaded - peer['downloaded']
+                    peer['downloaded'] = downloaded
+
                 if auth:
                     self.delete_peer(infohash,myid)
         
         elif not peer:
             ts[myid] = clock()
             peer = { 'ip': ip, 'port': port, 'left': left,
+                     'uploaded': uploaded, 'downloaded': downloaded,
                      'supportcrypto': supportcrypto,
                      'requirecrypto': requirecrypto }
             if mykey:
                 peer['key'] = mykey
             if gip:
                 peer['given ip'] = gip
+            if name:
+                peer['name'] = name
+                safe_name = clean_name(name)
+                if safe_name and safe_name not in torrent_name:
+                    torrent_name.append(safe_name)
             if port:
                 if not self.natcheck or islocal:
                     peer['nat'] = 0
@@ -945,7 +1082,7 @@ class Tracker:
             else:
                 peer['nat'] = 2**30
             if event == 'completed':
-                self.completed[infohash] += 1
+                stats['completed'] += 1
             if not left:
                 self.seedcount[infohash] += 1
                 
@@ -957,7 +1094,7 @@ class Tracker:
 
             ts[myid] = clock()
             if not left and peer['left']:
-                self.completed[infohash] += 1
+                stats['completed'] += 1
                 self.seedcount[infohash] += 1
                 if not peer.get('nat', -1):
                     for bc in self.becache[infohash]:
@@ -965,7 +1102,7 @@ class Tracker:
                             bc[1][myid] = bc[0][myid]
                             del bc[0][myid]
             elif left and not peer['left']:
-                self.completed[infohash] -= 1
+                stats['completed'] -= 1
                 self.seedcount[infohash] -= 1
                 if not peer.get('nat', -1):
                     for bc in self.becache[infohash]:
@@ -973,6 +1110,13 @@ class Tracker:
                             bc[0][myid] = bc[1][myid]
                             del bc[1][myid]
             peer['left'] = left
+            
+            if uploaded > peer.setdefault('uploaded', uploaded):
+                stats['uploaded'] += uploaded - peer['uploaded']
+                peer['uploaded'] = uploaded
+            if downloaded > peer.setdefault('downloaded', downloaded):
+                stats['downloaded'] += downloaded - peer['downloaded']
+                peer['downloaded'] = downloaded
 
             if port:
                 recheck = False
@@ -1559,8 +1703,16 @@ def size_format(s):
     
     if (s < 1024):
         r = str(s) + 'B'
+    elif (s < 10485):
+        r = str(int((s/1024.0)*100.0)/100.0) + 'KiB'
+    elif (s < 104857):
+        r = str(int((s/1024.0)*10.0)/10.0) + 'KiB'
     elif (s < 1048576):
         r = str(int(s/1024)) + 'KiB'
+    elif (s < 10737418L):
+        r = str(int((s/1048576.0)*100.0)/100.0) + 'MiB'
+    elif (s < 107374182L):
+        r = str(int((s/1048576.0)*10.0)/10.0) + 'MiB'
     elif (s < 1073741824L):
         r = str(int(s/1048576)) + 'MiB'
     elif (s < 1099511627776L):
