@@ -2,7 +2,7 @@
 # Modified by Cameron Dale
 # see LICENSE.txt for license information
 #
-# $Id: download_bt1.py 299 2007-10-17 03:48:40Z camrdale-guest $
+# $Id: download_bt1.py 381 2008-06-19 22:40:22Z camrdale-guest $
 
 """Manage a single download.
 
@@ -143,14 +143,19 @@ defaults = [
         'file contains IP range data in the format: xxx:xxx:ip1-ip2'),
     ('parse_ip_files', 60, 
         'seconds between reloading of allowed_ips and banned_ips lists for apt'),
-    # Backup HTTP Downloader
+    # Other Things
     ('disable_http_downloader', 0, 
         '(for testing purposes only) whether to disable the backup HTTP downloader'),
-    # Other Things
-    ('separate_all',0, 'whether to separate the architecture:all packages, ' +
+    ('separate_all', 3, 'whether to separate the architecture:all packages, ' +
         '0 = don\'t separate, 1 = separate and run architecture:all, ' +
         '2 = separate and run all architectures but all, ' + 
-        '3 = separate and run both (not for btdownloadheadless)'),
+        '3 = separate and run both'),
+    ('pieces_url', 'merkel.debian.org/~camrdale/extrapieces/',
+        'the remote URL to use to fetch information on the splitting of ' +
+        'large packages into pieces'),
+    ('long_lived_torrent_url', 'merkel.debian.org/~camrdale/uniquepieces/',
+        'the remote URL to use to fetch information on creating a long-lived ' +
+        'torrent for testing or unstable using unique piece numbers'),
     # End of Normal Options
     # BitTorrent Options
     ('keepalive_interval', 120.0,
@@ -187,6 +192,8 @@ defaults = [
     # Tracker Connections
     ('default_tracker', 'http://dttracker.debian.net:6969/announce', 
         'the default tracker address to use for new torrents'),
+    ('force_tracker', '', 
+        'set this to the tracker address to use for all torrents'),
     ('rerequest_interval', 5 * 60,
         'time to wait between requesting more peers'),
     ('min_peers', 20, 
@@ -426,8 +433,21 @@ def get_packages(file, url, separate_all = 0,
                 return (None, None)
 
         sub_pieces = getsubpieces(name)
-
-        (info, info_all) = getpieces(h, separate_all = separate_all, sub_pieces = sub_pieces)
+        
+        piece_ordering_all = {}
+        ordering_all_headers = {}
+        piece_ordering = {}
+        ordering_headers = {}
+        if separate_all:
+            (piece_ordering, ordering_headers) = getordering(name)
+            if separate_all in [1, 3]:
+                (piece_ordering_all, ordering_all_headers) = getordering(name, all = True)
+    
+        (info, info_all) = getpieces(h, separate_all = separate_all, sub_pieces = sub_pieces,
+                                     piece_ordering = piece_ordering,
+                                     piece_ordering_all = piece_ordering_all,
+                                     num_pieces = int(ordering_headers.get('NextPiece', 0)),
+                                     num_all_pieces = int(ordering_all_headers.get('NextPiece', 0)))
 
         response = None
         response_all = None
@@ -436,11 +456,29 @@ def get_packages(file, url, separate_all = 0,
                         'announce': default_tracker, 
                         'name': uniconvert(name)}
 
+            if "Tracker" in ordering_headers:
+                response['announce'] = ordering_headers["Tracker"].strip()
+                del ordering_headers["Tracker"]
+            if "Torrent" in ordering_headers:
+                response['identifier'] = binascii.a2b_hex(ordering_headers["Torrent"].strip())
+                del ordering_headers["Torrent"]
+            for header, value in ordering_headers.items():
+                response[header] = value.strip()
+
         if info_all:
             response_all = {'info': info_all,
                             'announce': default_tracker, 
                             'name': uniconvert(convert_all(name))}
     
+            if "Tracker" in ordering_all_headers:
+                response_all['announce'] = ordering_all_headers["Tracker"].strip()
+                del ordering_all_headers["Tracker"]
+            if "Torrent" in ordering_all_headers:
+                response_all['identifier'] = binascii.a2b_hex(ordering_all_headers["Torrent"].strip())
+                del ordering_all_headers["Torrent"]
+            for header, value in ordering_all_headers.items():
+                response_all[header] = value.strip()
+
     except IOError, e:
         logger.exception('problem getting Packages info')
         return (None, None)
@@ -487,6 +525,8 @@ class BT1Download:
     @ivar argslistheader: the header to print before the default config
     @type unpauseflag: C{threading.Event}
     @ivar unpauseflag: the flag to unset to pause the download
+    @type pickled_data: C{dictionary}
+    @ivar pickled_data: the saved data from a previous run
     @type downloader: L{BT1.Downloader.Downloader}
     @ivar downloader: the Downloader instance
     @type storagewrapper: L{BT1.StorageWrapper.StorageWrapper}
@@ -602,6 +642,7 @@ class BT1Download:
         self.argslistheader = argslistheader
         self.unpauseflag = Event()
         self.unpauseflag.set()
+        self.pickled_data = None
         self.downloader = None
         self.storagewrapper = None
         self.fileselector = None
@@ -724,10 +765,13 @@ class BT1Download:
 
                 files = []
                 for x in self.info['files']:
-                    n = file
-                    for i in x['path']:
-                        n = path.join(n, i)
-                    files.append((n, x['length']))
+                    if x['path']:
+                        n = file
+                        for i in x['path']:
+                            n = path.join(n, i)
+                        files.append((n, x['length']))
+                    else:
+                        files.append(('', 0L))
                     #Move directory creation to Storage
                     #make(n)
         except OSError, e:
@@ -805,30 +849,118 @@ class BT1Download:
             self.errorfunc(reason)
         
 
-    def find_files(self):
-        """Search through the save directory to find files which already exist.
+    def loadState(self):
+        """Load the save download state from a previous run.
         
         @rtype: C{list} of C{boolean}
-        @return: a list of which entries in the list L{files} already exist
+        @return: a list of the enabled files in the torrent
+            (first call only, subsequent calls will return None)
         
         """
+
+        # If it's already loaded, do nothing
+        if self.pickled_data is not None:
+            return
         
-        found_files = {}
-        for root, dirs, files in walk(self.filename):
-            for file in files:
-                found_files[path.join(root, file)] = 1
-        
-        if not found_files:
-            return None
-        
-        enabled_files = []
-        for file, length in self.files:
-            if file in found_files:
-                enabled_files.append(True)
+        try:
+            self.pickled_data = self.appdataobj.getTorrentData(self.infohash)
+        except:
+            logger.exception('Could not retrieve the pickled data')
+            self.pickled_data = None
+            
+        must_find_files = False
+
+        # Upgrade the saved state
+        if self.pickled_data and self.pickled_data.get('version', 0) < 1:
+            if 'resume data' in self.pickled_data:
+                new_priority = {}
+                if 'priority' in self.pickled_data['resume data']:
+                    try:
+                        old_priority = self.pickled_data['resume data']['priority']
+                        for i in xrange(min(len(old_priority), len(self.files))):
+                            if old_priority[i] >= 0:
+                                new_priority[self.files[i][0]] = old_priority[i]
+                        logger.info('Upgraded the old priority state')
+                    except:
+                        logger.exception('Previously saved priority state is corrupt, resetting')
+                        must_find_files = True
+                else:
+                    must_find_files = True
+                self.pickled_data['resume data']['priority'] = new_priority
+                new_files = {}
+                if 'files' in self.pickled_data['resume data']:
+                    try:
+                        old_files = self.pickled_data['resume data']['files']
+                        assert len(old_files) % 3 == 0
+                        old_files = [old_files[x:x+3] for x in xrange(0,len(old_files),3)]
+                        for i, size, mtime in old_files:
+                            if i < len(self.files):
+                                new_files[self.files[i][0]] = [size, mtime]
+                        logger.info('Upgraded the old saved files state')
+                    except:
+                        logger.exception('Previously saved file state is corrupt, resetting')
+                self.pickled_data['resume data']['files'] = new_files
+                if 'partial files' in self.pickled_data['resume data']:
+                    del self.pickled_data['resume data']['partial files']
             else:
-                enabled_files.append(False)
-        return enabled_files
+                self.pickled_data['resume data'] = {'priority': {}, 'files': {}}
+                must_find_files = True
+            self.pickled_data['stats'] = {'upload': 0L, 'download': 0L}
+            self.pickled_data['version'] = 1
+
+        # Initialize the saved state it if it wasn't found
+        if not self.pickled_data:
+            must_find_files = True
+            self.pickled_data = {}
+            self.pickled_data['resume data'] = {'priority': {}, 'files': {}}
+            self.pickled_data['stats'] = {'upload': 0L, 'download': 0L}
+            self.pickled_data['version'] = 1
+
+        enabled_files = []
+        found = 0
+        if must_find_files:
+            # No state, so search the download directory for old files to use
+            logger.info('no cached data, manually finding and hash checking old files')
+            self.pickled_data['resume data']['priority'] = {}
+            try:
+                logger.info('Searching %s for already downloaded files' % self.filename)
+                found_files = {}
+                for root, dirs, files in walk(self.filename):
+                    for file in files:
+                        found_files[path.join(root, file)] = 1
+                
+                if found_files:
+                    logger.info('Found %d downloaded files' % len(found_files.keys()))
+                    priority = {}
+                    for file, length in self.files:
+                        if file in found_files:
+                            found += 1
+                            priority[file] = 1
+                            enabled_files.append(True)
+                        else:
+                            enabled_files.append(False)
+                    logging.info('%d of the found files were also present in the %d files of the torrent' % (found, len(self.files)))
+                    self.pickled_data['resume data']['priority'] = priority
+                else:
+                    logger.info('No downloaded files were found')
+                
+            except:
+                logger.exception('Error occurred when manually finding the old files')
+                enabled_files = None
+        else:
+            # Have state, so build the list of enabled files
+            d = self.pickled_data['resume data']['priority']
+            for file, length in self.files:
+                if file in d and d[file] >= 0:
+                    found += 1
+                    enabled_files.append(True)
+                else:
+                    enabled_files.append(False)
+            logger.info('Of %d previous files, %d are still valid for this torrent' 
+                        % (len(d.keys()), found))
         
+        return enabled_files
+
     def initFiles(self, old_style = False):
         """Initialize the files for the download.
         
@@ -846,24 +978,7 @@ class BT1Download:
         if self.doneflag.isSet():
             return None
 
-        enabled_files = None
-        data = self.appdataobj.getTorrentData(self.infohash)
-        try:
-            d = data['resume data']['priority']
-            assert len(d) == len(self.files)
-            enabled_files = [x != -1 for x in d]
-        except:
-            if data:
-                logger.exception('pickled data is corrupt, manually finding and hash checking old files')
-            enabled_files = self.find_files()
-            if enabled_files:
-                priority = []
-                for found in enabled_files:
-                    if found:
-                        priority.append(1)
-                    else:
-                        priority.append(-1)
-                data = {'resume data': {'priority': priority}}
+        enabled_files = self.loadState()
 
         try:
             try:
@@ -896,10 +1011,7 @@ class BT1Download:
                                          self.storage, self.storagewrapper,
                                          self.rawserver.add_task, self.picker,
                                          self._failed)
-        if data:
-            data = data.get('resume data')
-            if data:
-                self.fileselector.unpickle(data)
+        self.fileselector.unpickle(self.pickled_data.get('resume data', {}))
 
         self.checking = True
         if old_style:
@@ -907,17 +1019,6 @@ class BT1Download:
         return self.storagewrapper.initialize
 
 
-    def getCachedTorrentData(self):
-        """Get the cached torrent data from the cache directory.
-        
-        @rtype: C{dictionary}
-        @return: the bdecoded cached data
-        
-        """
-        
-        return self.appdataobj.getTorrentData(self.infohash)
-
-
     def _make_upload(self, connection, ratelimiter, totalup):
         """Create a new Upload instance
         
@@ -1055,9 +1156,15 @@ class BT1Download:
         for i in xrange(self.len_pieces):
             if self.storagewrapper.do_I_have(i):
                 self.picker.complete(i)
+        
+        # Just to be sure
+        self.loadState()
+        
+        total_up = long(self.pickled_data['stats']['upload'])
+        total_down = long(self.pickled_data['stats']['download'])
         self.upmeasure = Measure(self.config['max_rate_period'],
-                            self.config['upload_rate_fudge'])
-        self.downmeasure = Measure(self.config['max_rate_period'])
+                            self.config['upload_rate_fudge'], saved_total = total_up)
+        self.downmeasure = Measure(self.config['max_rate_period'], saved_total = total_down)
 
         if ratelimiter:
             self.ratelimiter = ratelimiter
@@ -1160,7 +1267,8 @@ class BT1Download:
             self.storagewrapper.get_amount_left, 
             self.upmeasure.get_total, self.downmeasure.get_total,
             self.upmeasure.get_rate, self.downmeasure.get_rate,
-            self.doneflag, self.unpauseflag, seededfunc, force_rapid_update )
+            self.doneflag, self.unpauseflag, seededfunc, force_rapid_update,
+            self.response.get('name', ''))
 
         self.rerequest.start()
 
@@ -1216,6 +1324,9 @@ class BT1Download:
             self.storage.close()
             self.rerequest_stopped()
         if self.fileselector and self.started:
+            torrentdata['version'] = 1
+            torrentdata['stats'] = {'upload': self.upmeasure.get_total(),
+                                    'download': self.downmeasure.get_total()}
             if not self.failed:
                 self.fileselector.finish()
                 torrentdata['resume data'] = self.fileselector.pickle()
@@ -1446,6 +1557,21 @@ class BT1Download:
         except:
             return None
 
+    def getFilepath(self, f):
+        """Get the stored file name from a file index.
+        
+        @type f: C{int}
+        @param f: the index of the file
+        @rtype: C{string}
+        @return: the file name where the indexed file is stored.
+        
+        """
+        
+        try:
+            return self.files[f][0]
+        except:
+            return None
+
     def Pause(self):
         """Schedule the pausing of the download.
 
