@@ -2,7 +2,7 @@
 # Modified by Cameron Dale
 # see LICENSE.txt for license information
 #
-# $Id: launchmanycore.py 320 2007-11-07 00:06:50Z camrdale-guest $
+# $Id: launchmanycore.py 389 2008-06-22 19:23:06Z camrdale-guest $
 
 """Manage the downloading of multiple torrents in one process.
 
@@ -67,6 +67,8 @@ class SingleDownload:
     @ivar controller: the manager for all torrent downloads
     @type hash: C{string}
     @ivar hash: the info hash of the torrent
+    @type identifier: C{string}
+    @ivar identifier: the identifier of the torrent
     @type response: C{dictionary}
     @ivar response: the meta info for the torrent
     @type config: C{dictionary}
@@ -103,13 +105,15 @@ class SingleDownload:
     
     """
     
-    def __init__(self, controller, hash, response, config, myid):
+    def __init__(self, controller, hash, identifier, response, config, myid):
         """Initialize the instance and start a new downloader.
         
         @type controller: L{LaunchMany}
         @param controller: the manager for all torrent downloads
         @type hash: C{string}
         @param hash: the info hash of the torrent
+        @type identifier: C{string}
+        @param identifier: the identifier of the torrent
         @type response: C{dictionary}
         @param response: the meta info for the torrent
         @type config: C{dictionary}
@@ -121,6 +125,7 @@ class SingleDownload:
         
         self.controller = controller
         self.hash = hash
+        self.identifier = identifier
         self.response = response
         self.config = config
         
@@ -136,11 +141,11 @@ class SingleDownload:
         self.status_errtime = 0
         self.status_done = 0.0
 
-        self.rawserver = controller.handler.newRawServer(hash, self.doneflag)
+        self.rawserver = controller.handler.newRawServer(hash, identifier, self.doneflag)
 
         d = BT1Download(self.update_status, self.finished, self.error,
-                        self.doneflag, config, response,
-                        hash, myid, self.rawserver, controller.listen_port, 
+                        self.doneflag, config, response, identifier,
+                        myid, self.rawserver, controller.listen_port, 
                         self.controller.configdir)
         self.d = d
 
@@ -428,6 +433,7 @@ class LaunchMany:
         data = []
         for hash in self.torrent_list:
             cache = self.torrent_cache[hash]
+            id = cache['metainfo'].get('identifier', hash)
             if self.config['display_path']:
                 name = cache['path']
             else:
@@ -484,7 +490,7 @@ class LaunchMany:
             else:
                 msg = ''
 
-            data.append(( name, hash, status, progress, peers, seeds, seedsmsg, 
+            data.append(( name, id, status, progress, peers, seeds, seedsmsg, 
                           dist, uprate, dnrate, upamt, dnamt, size, t, msg ))
 
         return data
@@ -537,7 +543,9 @@ class LaunchMany:
             # Stop any running previous versions of the same torrent
             same_names = []
             for old_hash in self.torrent_list:
-                if self.torrent_cache[old_hash]['name'] == data['name']:
+                if (self.torrent_cache[old_hash]['name'] == data['name'] or
+                    self.torrent_cache[old_hash]['metainfo'].get('identifier', old_hash) == 
+                        data['metainfo'].get('identifier', hash)):
                     same_names.append(old_hash)
             for old_hash in same_names:
                 self.remove(old_hash)
@@ -556,7 +564,15 @@ class LaunchMany:
             logger.error('Asked to start a torrent that is not in the cache')
             return
         
-        logger.info('Starting torrent: '+str(binascii.b2a_hex(hash)))
+        # Assign the torrent identifier from the metainfo data
+        if "identifier" in self.torrent_cache[hash]['metainfo']:
+            id = self.torrent_cache[hash]['metainfo']['identifier']
+            logger.info('Starting torrent: '+str(binascii.b2a_hex(hash)) + 
+                        ' identified by: '+str(binascii.b2a_hex(id)))
+        else:
+            id = hash
+            logger.info('Starting torrent: '+str(binascii.b2a_hex(hash)))
+
         c = self.counter
         self.counter += 1
         x = ''
@@ -564,7 +580,8 @@ class LaunchMany:
             x = mapbase64[c & 0x3F]+x
             c >>= 6
         peer_id = createPeerID(x)
-        d = SingleDownload(self, hash, self.torrent_cache[hash]['metainfo'], self.config, peer_id)
+        d = SingleDownload(self, hash, id, self.torrent_cache[hash]['metainfo'],
+                           self.config, peer_id)
         self.torrent_list.append(hash)
         self.downloads[hash] = d
         d.start()
@@ -648,7 +665,7 @@ class LaunchMany:
                 file_num += 1
                 
                 # Check that the file ends with the desired file name
-                if file.endswith('/'.join(f['path'])):
+                if f['path'] and file.endswith('/'.join(f['path'])):
                     logger.debug('Found file in: '+str(binascii.b2a_hex(hash)))
                     found_torrents.append((hash, file_num))
         
